/*
	LablPCRE - PCRE[posix] binding for Objective Caml

	author: Robert Roessler
	email: robertr@rftp.com
	web: http://www.rftp.com

--------------------------------------------------------------------------------
Copyright (c) 2005-2007 Robert Roessler
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

    * Redistributions of source code must retain the above copyright notice,
      this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of Robert Roessler nor the names of its contributors
      may be used to endorse or promote products derived from this
      software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.
--------------------------------------------------------------------------------
*/

/* ml_pcre.cpp - interface from PCRE OCaml code -> C-land */

#include <string.h>
#include "pcreposix.h"
#include "pcre.h"

extern "C" {
#include <caml/memory.h>
#include <caml/alloc.h>
#include <caml/custom.h>
#include <caml/fail.h>

//	release PCRE-allocated resources
static void finalize(value v)
{
	regfree((regex_t*)Data_custom_val(v));
}

//	compile a [posix] regex pattern
CAMLprim value ml_regcomp(value pattern, value flags)
{
	CAMLparam2(pattern, flags);
	CAMLlocal1(re_block);
	// compile regex to local temp (avoiding alloc in case of error)
	regex_t re;
	const int r = regcomp(&re, String_val(pattern), Int_val(flags));
	// throw Failure exception if error...
	if (r != 0) {
		char buf[256];
		regerror(r, NULL, buf, sizeof buf);
		caml_failwith(buf);
	}
	// ... otherwise, alloc custom block for regex_t struct...
	static custom_operations co = {
		"http://www.rftp.com/LablPCRE/regex_t",
		finalize,
		custom_compare_default,
		custom_hash_default,
		custom_serialize_default, custom_deserialize_default
	};
	re_block = caml_alloc_custom(&co, sizeof(regex_t), 0, 1);
	// ... and copy from local temp
	memcpy(Data_custom_val(re_block), &re, sizeof(regex_t));
	CAMLreturn(re_block);
}

//	execute a compiled [posix] regex pattern (FULL)
CAMLprim value ml_regexec(value re_block, value target, value n_match, value pos, value flags)
{
	CAMLparam5(re_block, target, n_match, pos, flags);
	CAMLlocal2(rm_block, rv_block);
	const int n_m = Int_val(n_match);
	// alloc custom block for regmatch_t structs AS REQUIRED
	if (n_m) {
		static custom_operations co = {
			"http://www.rftp.com/LablPCRE/regmatch_t",
			NULL,
			custom_compare_default,
			custom_hash_default,
			custom_serialize_default, custom_deserialize_default
		};
		// N.B. - this will be discarded if regexec fails (returns NON-zero)
		rm_block = caml_alloc_custom(&co, sizeof(regmatch_t) * n_m, 0, 1);
	}
	const int off = Int_val(pos);
	// execute compiled regex pattern
	const int r = regexec((regex_t*)Data_custom_val(re_block),
		String_val(target) + off, n_m,
		(regmatch_t*)(n_m? Data_custom_val(rm_block): NULL), Int_val(flags));
	// create return value
	if (r != 0 || n_m == 0)
		// "Error of regerror"
		rv_block = caml_alloc_small(1, 0), Field(rv_block, 0) = Val_int(r);
	else {
		// perform offset "adjustments" AS REQUIRED
		if (off > 0) {
			regmatch_t* rm = (regmatch_t*)Data_custom_val(rm_block);
			for (int i = 0; i < n_m; i++)
				// (ONLY fix up used entries)
				if (rm[i].rm_so >= 0)
					rm[i].rm_so += off, rm[i].rm_eo += off;
		}
		// "Match of regmatch"
		rv_block = caml_alloc_small(1, 1), Field(rv_block, 0) = rm_block;
	}
	CAMLreturn(rv_block);
}

//	execute a compiled [posix] regex pattern (MATCH ONLY)
CAMLprim value ml_regmatch(value re_block, value target, value pos, value flags)
{
	CAMLparam4(re_block, target, pos, flags);
	// execute compiled regex pattern
	const value r = regexec((regex_t*)Data_custom_val(re_block),
		String_val(target) + Int_val(pos), 0, NULL, Int_val(flags))? Val_false: Val_true;
	CAMLreturn(r);
}

//	retrieve a [posix] error string
CAMLprim value ml_regerror(value error, value re_block_opt)
{
	CAMLparam2(error, re_block_opt);
	CAMLlocal1(s);
	// get the [posix] error string
	char buf[256];
	regerror(Int_val(error), (regex_t*)((re_block_opt != Val_unit)?
		Data_custom_val(Field(re_block_opt, 0)): NULL), buf, sizeof buf);
	// allocate a [string] return value and copy (regerror forces termination)
	s = caml_copy_string(buf);
	CAMLreturn(s);
}

//	retrieve a match component from a regexec result
CAMLprim value ml_sub(value rm_block, value target, value i_val)
{
	CAMLparam3(rm_block, target, i_val);
	CAMLlocal1(s);
	const int i = Int_val(i_val);
	const int n_m = (Bosize_val(rm_block) - sizeof(value)) / sizeof(regmatch_t);
	if (i < 0 || i >= n_m)
		caml_invalid_argument("Pcre.ml_sub");
	// fetch this BEFORE any allocation operations
	const regmatch_t rm = ((regmatch_t*)Data_custom_val(rm_block))[i];
	const int len = rm.rm_eo - rm.rm_so;
	// allocate a [string] return value and copy AS REQUIRED
	// N.B. - unused match entries (-1, -1) will have len == 0
	s = caml_alloc_string(len);
	memcpy(String_val(s), String_val(target) + rm.rm_so, len);
	CAMLreturn(s);
}

//	retrieve match component info from a regexec result
CAMLprim value ml_sub_start(value rm_block, value i_val)
{
	CAMLparam2(rm_block, i_val);
	const int i = Int_val(i_val);
	const int n_m = (Bosize_val(rm_block) - sizeof(value)) / sizeof(regmatch_t);
	if (i < 0 || i >= n_m)
		caml_invalid_argument("Pcre.ml_sub_start");
	const regmatch_t rm = ((regmatch_t*)Data_custom_val(rm_block))[i];
	CAMLreturn(Val_int(rm.rm_so));
}

CAMLprim value ml_sub_end(value rm_block, value i_val)
{
	CAMLparam2(rm_block, i_val);
	const int i = Int_val(i_val);
	const int n_m = (Bosize_val(rm_block) - sizeof(value)) / sizeof(regmatch_t);
	if (i < 0 || i >= n_m)
		caml_invalid_argument("Pcre.ml_sub_end");
	const regmatch_t rm = ((regmatch_t*)Data_custom_val(rm_block))[i];
	CAMLreturn(Val_int(rm.rm_eo));
}

//	translate [posix] compile options (as OCaml sum type) -> [posix] int flags
CAMLprim value ml_int_of_c_flag(value c_flag)
{
	static const int t[] = {REG_ICASE, REG_NEWLINE, REG_DOTALL, REG_EXTENDED, REG_NOSUB};
	return Val_int(t[Int_val(c_flag)]);
}

//	translate [posix] execution options (as OCaml sum type) -> [posix] int flags
CAMLprim value ml_int_of_x_flag(value x_flag)
{
	static const int t[] = {REG_NOTBOL, REG_NOTEOL};
	return Val_int(t[Int_val(x_flag)]);
}

}
