(*
	LablPCRE - PCRE[posix] binding for Objective Caml

	author: Robert Roessler
	email: robertr@rftp.com
	web: http://www.rftp.com

--------------------------------------------------------------------------------
Copyright (c) 2005,2006 Robert Roessler
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

    * Redistributions of source code must retain the above copyright notice,
      this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of Robert Roessler nor the names of its contributors
      may be used to endorse or promote products derived from this
      software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.
--------------------------------------------------------------------------------
*)

(* pcre.ml - implementation of PCRE[posix] binding for OCaml *)

(* standard PCRE[posix] compiled regexp and match results array *)
type regexp (* regex_t in C-land *)
type regmatch (* regmatch_t in C-land *)

(* standard PCRE[posix] error values
	N.B. - REG_NOERROR is something of a hack, but since the "official"
	enum in pcreposix.h begins with "REG_ASSERT=1"... it works *)
type regerror = REG_NOERROR
	| REG_ASSERT | REG_BADBR | REG_BADPAT | REG_BADRPT | REG_EBRACE
	| REG_EBRACK | REG_ECOLLATE | REG_ECTYPE | REG_EESCAPE | REG_EMPTY
	| REG_EPAREN | REG_ERANGE | REG_ESIZE | REG_ESPACE | REG_ESUBREG
	| REG_INVARG | REG_NOMATCH

(* type to hold PCRE[posix] match and/or error state *)
type t = {matches:regmatch option; target:string option; n_match:int; error:regerror}

(* result of [low-level] regexec_raw *)
type regresult_raw = Error of regerror | Match of regmatch

(* interface functions for accessing PCRE API and data structs *)
external regcomp_raw: string -> int -> regexp = "ml_regcomp"
external regexec_raw: regexp -> string -> int -> int -> int -> regresult_raw = "ml_regexec"
external regmatch_raw: regexp -> string -> int -> int -> bool = "ml_regmatch" "noalloc"
external regerror_raw: regerror -> regexp option -> string = "ml_regerror"
external sub_raw: regmatch -> string -> int -> string = "ml_sub"
external sub_start_raw: regmatch -> int -> int = "ml_sub_start"
external sub_end_raw: regmatch -> int -> int = "ml_sub_end"

(* standard PCRE[posix] regexp compile flags *)
type c_flag = REG_ICASE | REG_NEWLINE | REG_DOTALL | REG_EXTENDED | REG_NOSUB

(* standard PCRE[posix] regexp execute flags *)
type x_flag = REG_NOTBOL | REG_NOTEOL

(* interface functions for translating PCRE flags *)
external int_of_c_flag: c_flag -> int = "ml_int_of_c_flag"
external int_of_x_flag: x_flag -> int = "ml_int_of_x_flag"

(* OR int flags together *)
(* special-case 0 and 1 element cases, since folds are expensive(?) *)
let or_all f = function
	| [] -> 0
	| h :: [] -> f h
	| l -> List.fold_left (fun o e -> o lor (f e)) 0 l

(* compile a PCRE[posix] regexp *)
let regcomp ?(flags=[]) pat =
	regcomp_raw pat (or_all int_of_c_flag flags)

(* needless to say, this will NOT appear in exported interface *)
let make ?(matches=None) ?(target=None) ?(n_match=0) error =
	{matches=matches; target=target; n_match=n_match; error=error}

(* execute a PCRE[posix] compiled regexp (FULL) *)
let regexec ~reg ?(n_match=0) ?(pos=0) ?(flags=[]) target =
	if pos < 0 || pos > String.length target then
		invalid_arg "Pcre.regexec";
	match regexec_raw reg target n_match pos (or_all int_of_x_flag flags) with
	| Error e -> make e
	| Match m -> make REG_NOERROR
		~matches:(Some m) ~target:(Some target) ~n_match

(* execute a PCRE[posix] compiled regexp (MATCH ONLY) *)
let regmatch ~reg ?(pos=0) ?(flags=[]) target =
	regmatch_raw reg target pos (or_all int_of_x_flag flags)

(* retrieve PCRE[posix] error string *)
let regerror ?reg error = regerror_raw error reg

let success r = r.error == REG_NOERROR

let error r = r.error

let error_string r = regerror r.error

let length r = r.n_match

let sub r i =
	match r.matches, r.target with
	| Some m, Some t -> sub_raw m t i
	| _ -> failwith "Pcre.sub"

let sub_start r i =
	match r.matches with
	| Some m -> sub_start_raw m i
	| _ -> failwith "Pcre.sub_start"

let sub_end r i =
	match r.matches with
	| Some m -> sub_end_raw m i
	| _ -> failwith "Pcre.sub_end"

let exists r i = sub_start r i >= 0

(* object to hold PCRE[posix] match and/or error type ("old" interface) *)
class regresult r = object (self)
	val r = r
	method success = success r
	method error = error r
	method length = length r
	method substring = sub r
	method substring_start = sub_start r
	method substring_end = sub_end r
    method substring_captured = exists r
end
